"""
The Slogan Log Analyser

Purpose: Hierarchical storage and interactive analysis of web server log files
Version: 0.05 <dogfood>
Author: Andrew Clover <aclover@1value.com>
Licence: GPL <http://www.gnu.org/copyleft/gpl.html>
"""

## Imports
## =============================================================================

import os, os.path, copy, string, re, time

__all__= ['html', 'languages', 'databases', 'traits']

## autoPackage

# Languages, databases and traits are detected automatically when they are
# dropped into the slogan package. This is a function to help do it.

def autoPackage(path):
  all= []
  for file in os.listdir(os.path.dirname(path)):
    name, extension= os.path.splitext(file)
    if extension[1:]=='py' and name!='__init__':
      all.append(name)
  return all

import slogan.languages, slogan.databases, slogan.traits

## Initialisation
## =============================================================================

false, true= range(2)

## idDict

# take a list of classes and make a dictionary of them, using their 'identity'
# property as a key.

def idDict(all):
  dict= {}
  for item in all:
    dict[item.identity]= item
  return dict

# Shortcut names for standard logfile formats

logLineCommon= [
  ('&host; &ignore; &ignore; [&date;:&time;] "&method; &file; &ignore;" &status; &ignore;', true)
]

logLineCombined= [
  ('&host; &ignore; &cookie; [&date;:&time; &ignore;] "&method; &file; &ignore;" &status; &ignore; &"referrer; &"browser;', true)
]

logLineExtended= [
  ('#Fields: &extended;', false),
  ('#Date: &date;', false),
  ('#&ignore;', false)
]

## config

## transformFormatString

# convert Slogan's own style of format string, which uses HTML-style &entities;
# into a regexp object which will match a string in this format, and a list of
# traits that correspond to each group in the regexp match.

findAttributes= re.compile('(&[^;]*;)')
quoteEncoded= '(-|(?:"(?:[^"]|(?:""))*"))'
notEncoded= '(.*?)'
def transformFormatString((format, makehit)):
  lineRegexp, lineTraits, lineEncodings= '', [], []
  for token in findAttributes.split(format):
    if token[:1]=='&' and token[-1:]==';':
      entity= token[1:-1]
      if entity=='amp':
        lineRegexp= lineRegexp+'&'
      else:
        encoding= None
        if entity[:1] in ['"', '+']:
          encoding, entity= entity[:1], entity[1:]
        if not traitObjects.has_key(entity):
          raise NameError, 'Trait "'+entity+'" not known'
        lineTraits.append(traitObjects[entity])
        lineEncodings.append(encoding)
        lineRegexp= lineRegexp+[notEncoded, quoteEncoded][encoding=='"']
    else:
      lineRegexp= lineRegexp+re.escape(token)
  return (re.compile(lineRegexp+'$'), lineTraits, lineEncodings, makehit)

## initialise

# initialise data structures taking into account settings for the specified setup

def initialise(set):
  global config, database, language, traitObjects, traitPriorities
  global logNameFormat, logLineFormats, logLineLength
  import slogan.settings
  config= slogan.settings.settings[set][0]

# Get database connection and language objects from whichever file needs to
# be imported to get them.

  if config['language'] in slogan.languages.__all__:
    __import__('slogan.languages.'+config['language'])
    language= getattr(slogan.languages, config['language']).Language()
  else:
    raise NameError, 'Language "'+config['language']+'" not known'
  if config['dbDriver'] in slogan.databases.__all__:
    __import__('slogan.databases.'+config['dbDriver'])
    database= getattr(slogan.databases, config['dbDriver']).Driver()
  else:
    raise NameError, 'Database driver "'+config['dbDriver']+'" not known'
  database.connect(
    config['dbHost'], config['dbName'], config['dbUser'], config['dbPass']
  )

# Instantiate all traits

  traitObjects= {}
  traitPriorities= []
  for trait in slogan.traits.__all__:
    __import__('slogan.traits.'+trait)
    traitObjects[trait]= getattr(slogan.traits, trait).Trait()
    priorityIndex= 0
    while len(traitPriorities)>priorityIndex and traitObjects[traitPriorities[priorityIndex]].priority>=traitObjects[trait].priority:
      priorityIndex= priorityIndex+1
    traitPriorities[priorityIndex:priorityIndex]= [trait]

# Each possible logLine format string must be processed into a regexp and
# trait list, whether or not it will cause a hit in the log database. The
# logName string has to be processed in the same way too.

# Also, a length is held for the processed logLine list. This is because further
# elements may be added to the list during the parsing of the logfile, and it
# must be possible to remove them again. In particular this is used by the
# ExtendedTrait parser to add one new Extended format line at a time to the
# list.

  logNameFormat= transformFormatString((config['logName'], false))
  logLineFormats= map(transformFormatString, config['logLine'])
  logLineLength= len(logLineFormats)

## SQL commands
## =============================================================================

## sqlInsert

# Take a dictionary of {trait: [values]} and convert into an SQL statement that
# will INSERT the data into the database as bits of flat text.

def sqlInsert(dict):
  rows= []
  vals= []
  for trait in traitObjects.values():
    i= 0
    while i<len(dict.get(trait.identity, [])):
      rows.append(trait.identity+str(i))
      vals.append(dict[trait.identity][i])
      i= i+1
    while i<len(trait.size):
      rows.append(trait.identity+str(i))
      vals.append('')
      i= i+1
  rows= string.join(rows, ', ')
  vals= string.join(map(database.sqlLiteral, vals), ', ')
  database.command('INSERT INTO sloganlogs ('+rows+') VALUES ('+vals+');')

## sqlSelect

# Take a dictionary of [column: value] conditions, a column to read possible
# values of, and a type of sorting to apply, and construct an SQL SELECT
# statement to count the number of hits for each possible value.

# Return the results as a list of tuples (value, hits), sorted either
# asciibetically or by hits. Use a negative sortType to get fewest-hits-first.

# The length of the tuple and the number of sort types may be expanded to
# accomodate counting of bytes served or time taken.

sortType_ascii, sortType_hits= range(2)

def sqlSelect(conditions, column, sortType):
  global lamarg
  sql= 'SELECT '+column+' AS v, COUNT(*) AS h FROM sloganlogs'
  if conditions:
    lamarg= conditions
    sqlCondition= lambda key: key+'='+database.sqlLiteral(lamarg[key])
    sql=sql+' WHERE '+string.join(map(sqlCondition, conditions.keys()), ' AND ')
  sql= sql+' GROUP BY '+column
  sql= sql+' ORDER BY '+['v','h'][abs(sortType)]+[' ASC',' DESC'][sortType>0]+';'
  rows= database.query(sql)
  return map(lambda row: (row['v'], row['h']), rows)

## sqlCreate

# Make the 'sloganlogs' table with the sizes specified in the trait classes.
# Also create the 'sloganfiles' table to store how far files have been parsed.

def sqlCreate():
  global lamarg
  fields= []
  indices= []
  for trait in traitObjects.values():
    for i in range(len(trait.size)):
      fields.append(
        trait.identity+str(i)+' '+(database.sqlTextType % trait.size[i])
      )
    if len(trait.size)>0:
      lamarg= trait.identity
      cols= string.join(
        map(lambda i: lamarg+str(i), range(len(trait.size))), ', '
      )
      indices.append(
        'CREATE INDEX '+trait.identity+'x ON sloganlogs ('+cols+');'
      )
  database.command('CREATE TABLE sloganlogs ('+string.join(fields, ', ')+');')
  for index in indices:
    database.command(index)
  sql= 'CREATE TABLE sloganfiles (file0 '+(database.sqlTextType % 255)+', '
  sql= sql+'size0 '+(database.sqlTextType % 12)+', '
  sql= sql+'time0 '+(database.sqlTextType % 12)+');'
  database.command(sql)

## sqlDelete

# Just delete the sloganlogs table. Deletion followed by recreation is the only
# reliable way of changing the table format for sloganlogs, since ALTER
# TABLE is so spottily supported.

# sloganfiles is also deleted since sqlCreate() will re-make it, and the
# information in it makes no sense if sloganlogs does not exist.

def sqlDelete():
  database.command('DROP TABLE sloganlogs;')
  database.command('DROP TABLE sloganfiles;')

## Logfile parsing
## =============================================================================

## updateTraits

# updateTraits takes a regexp match object, decodes each captured part as
# appropriate and passes each part to the relevant trait class to parse. It then
# normalises the results to match the subpart list length given in the trait
# class's size property, and trims any subparts that are too long according to
# the size property. The final results are then stored in the dictionary that is
# passed to updateTraits.

def updateTraits(dict, match, order, encodings):
  for i in range(len(order)):
    trait= order[i]
    logpart= match.group(i+1)
    if logpart=='-':
      logpart= ''
    else:
      if encodings[i]=='"':
        if logpart[:1]=='"' and logpart[-1:]=='"':
          logpart= logpart[1:-1]
        logpart= string.replace(logpart, '""', '"')
      else:
        if encodings[i]=='+':
          logpart= string.replace(logpart, '+', ' ')
    values= trait.read(logpart)
    if len(trait.size)==0:
      continue
    values= values[:len(trait.size)]
    for j in range(len(values)):
      if len(values[j])>trait.size[j]:
        values[j]= values[j][:trait.size[j]]
    while len(values)<len(trait.size):
      values.append('')
    dict[order[i].identity]= values

## readLogfile

# readLogfile opens a file and reads data from it, using the dictionary of
# trait values passed to it. Thus it can be passed trait values from the name
# of the file. Only lines which begin from a seek position higher than
# bytesAlreadyProcessed are entered into the database.

# If logmanagement is set to autoseek, we can skip the already-done log lines
# completely to save time, because no previous lines can affect the trait values
# of later lines. Otherwise we have to read through the whole file again.
# Extended format logfiles, after all, may change the format of a line at any
# point. The ignoreHits flag is set true when this is happening.

# The function returns the seek pointer of the end of the file.

def readLogfile(filename, current, bytesAlreadyProcessed):
  log= open(filename, 'rb')
  ignoreHits= false
  if bytesAlreadyProcessed!=0:
    if config['logManagement']=='autoseek':
      log.seek(bytesAlreadyProcessed-1)
      while true:
        c= log.read(1)
        if c=='' or c=='\x0A':
          break
    else:
      ignoreHits= true
  while true:
    line= log.readline()
    if line=='':
      break

# The file must be opened as binary for seeks to be reliable, so we must also
# detect and remove two different types of line ending. Having done this we
# attempt to match each possible log line format in turn until one works or
# we've tried them all. On a match, the currently trait dictionary is updated
# with the new values, and said dictionary entered into the database if
# the line represents a hit.

    if line[-2:]=='\x0D\x0A':
      line= line[:-2]
    else:
      if line[-1:]=='\x0A':
        line= line[:-1]
    for (format, order, encodings, makehit) in logLineFormats:
      match= format.match(line)
      if match!=None:
        updateTraits(current, match, order, encodings)
        if ignoreHits:
          if log.tell()>=bytesAlreadyProcessed:
            ignoreHits= false
        else:
          if makehit:
            dict= {}
            for trait in current.keys():
              for i in range(len(current[trait])):
                dict[trait+str(i)]= current[trait][i]
            sqlInsert(current)
        break
  fileEnd= log.tell()
  log.close()
  return fileEnd

## Entry points
## =============================================================================

## updateDatabase

# entry point to Slogan's logfile reading bits. First it finds the names of
# eligable logfiles, and any traits embedded in their filenames.

def updateDatabase():
  if os.path.isfile(config['logPath']):
    nameTraits= {}
    match= logNameFormat[0].match(os.path.basename(config['logPath']))
    if match!=None:
      updateTraits(nameTraits, match, logNameFormat[1], logNameFormat[2])
    files= [(config['logPath'], nameTraits)]
  else:
    if os.path.isdir(config['logPath']):
      files= []
      for file in os.listdir(config['logPath']):
        path= os.path.join(config['logPath'], file)
        match= logNameFormat[0].match(file)
        if match!=None:
          nameTraits= {}
          updateTraits(nameTraits, match, logNameFormat[1], logNameFormat[2])
          files.append(path, nameTraits)

    else:
      raise IOError, 'No such file or directory, logPath: '+config['logPath']

# Next, log management policy dictates how much of the files are read. In auto
# mode, file sizes are compared to what they were last time they were read into
# the database, and the files are only processed if they have grown. The
# position reading stopped at last time is passed to readLogfile so it can
# skip some bits and that.

  if config['logManagement']=='reload':
    database.command('DELETE FROM sloganlogs;')
  for (path, initial) in files:
    now= time.time()
    if config['logManagement'] in ['auto', 'autoseek']:
      sql='SELECT * FROM sloganfiles WHERE file0='+database.sqlLiteral(path)+';'
      rows= database.query(sql)
      if len(rows)>0:
        start= int(rows[0]['size0'])
        if start<os.stat(path)[6]:
          end= readLogfile(path, initial, start)
          sql= 'UPDATE sloganfiles SET size0='+database.sqlLiteral(str(end))
          sql= sql+', time0='+database.sqlLiteral(str(int(now)))
          sql= sql+' WHERE file0='+database.sqlLiteral(path)+';'
          database.command(sql)
        continue;
    end= readLogfile(path, initial, 0)
    if config['logManagement'] in ['auto', 'autoseek']:
      sql= 'INSERT INTO sloganfiles (file0, size0, time0) VALUES ('
      sql= sql+database.sqlLiteral(path)+', '+database.sqlLiteral(str(end))
      sql= sql+', '+database.sqlLiteral(str(int(now)))+');'
      database.command(sql)

## query

# main entry point for log queries. Take a list of (trait: [values]) tuples,
# where each tuple represents the condition that the specified trait must have
# the given values up to the length of the values list.

# A list of tuples (child, hits) is returned for each possible child value of
# the last condition in the list. The list is sorted according to the sortType
# as given in sqlSelect. If there can be no child values because the last
# trait in the list is already fully specified, ('', count) is returned.

def query(conditions, sortType):
  dict= {}
  for (trait, values) in conditions:
    for i in range(min(len(values), len(traitObjects[trait].size))):
      dict[trait+str(i)]= values[i]
  if len(traitObjects[conditions[-1][0]].size)>len(conditions[-1][1]):
    return sqlSelect(dict, conditions[-1][0]+str(len(conditions[-1][1])), sortType)
  else:
    return map(lambda (v, h): ('', h), sqlSelect(dict, conditions[-1][0]+str(len(conditions[-1][1])-1), sortType))

